# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.

""" Helper functions and classes to deal with files and dirs and stuff

Featuring::

* the cd context manager pwd(); with cd(other_dir): pwd();   pwd();
  will print current working dir, then other_dir, then first dir again
* a wrapper around "df" to determine size and usage of file systems

.. todo:: test and change default method for get_filesystem_fill_states

.. codeauthor:: Intra2net
"""

from __future__ import print_function

from contextlib import contextmanager
import os
import os.path
from warnings import warn
from math import floor, ceil
import re

from call_helpers import call_and_capture
from iter_helpers import pairwise


@contextmanager
def cd(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    Taken from comment for python recipe by Greg Warner at
    http://code.activestate.com/recipes/576620-changedirectory-context-manager/
    (MIT license)
    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


DF_CMD = ['/usr/bin/df', '--no-sync', '--portability']
DF_SIZE_UNIT = 1024
FS_FILL_METHOD_DF = 'df'
FS_FILL_METHOD_STATVFS = 'statvfs'

#: types of mounts found in /proc/mounts to be ignored in get_all_mounts
IGNORE_MOUNT_TYPES = 'cgroup', 'pstore'


class FilesystemFillState:
    """ representation of 1 line of the 'df' command

    has fields filesystem, size, used, available, capacity, mount_point

    Note that only apprixomately capacity == used/size
     and that only approximately used + available == size
     and that all sizes are in bytes
    """

    def __init__(self):
        self.name = None
        self.size = None
        self.used = None
        self.available = None
        self.capacity = None
        self.mount_point = None

    def __str__(self):
        return '[Filesystem {0} mounted at {1}: {2}% used]' \
               .format(self.name, self.mount_point, self.capacity)


def get_filesystem_fill_states(method=FS_FILL_METHOD_DF, *args, **kwargs):
    """ get fill state on all filesystems

    :param method: FS_FILL_METHOD_DF (use df_wrapper, forks) or
                   FS_FILL_METHOD_STATVFS (uses get_all_statvfs)
    all other args and kwargs are forwarded
    """

    if method == FS_FILL_METHOD_DF:
        return df_wrapper(*args, **kwargs)
    else:
        return get_all_statvfs_fills(*args, **kwargs)


def df_wrapper():
    """ parses the output of cmd 'df'

    :returns: list of  :py:class:`FilesystemFillState`

    Tries to give results compatible with those from get_filesystem_fill_states
    """

    # call
    code, out, err = call_and_capture(DF_CMD)

    # warn if unexpected outcome
    if code != 0:
        warn('df returned non-zero exit code {0}!'.format(code))
    if err:
        for line in err:
            warn('df produced output to stderr: "{0}"'.format(line))

    # find columns in output that are just spaces
    min_len = min(len(line) for line in out)
    separator_cols = [idx for idx in range(min_len) if \
                      all(line[idx] == ' ' for line in out)]
    checked_cols = [separator_cols[0], ]
    for prev_col, col in pairwise(separator_cols):
        if col != prev_col+1:
            checked_cols.append(col)
    separator_cols = checked_cols

    # check columns and their header
    if len(separator_cols) != 5:
        raise ValueError('unexpected number of separator columns: {0}'
                         .format(separator_cols))  # must eliminate neighbours?

    title_line = out[0]
    title = title_line[ : separator_cols[0]].strip()
    if title != 'Filesystem':
        warn('Unexpected column title: "{0}" != "Filesystem"!'
             .format(title))
    title = title_line[separator_cols[0] : separator_cols[1]].strip()
    if title != '1024-blocks':
        warn('Unexpected column title: "{0}" != "1024-blocks"!'
             .format(title))
    title = title_line[separator_cols[1] : separator_cols[2]].strip()
    if title != 'Used':
        warn('Unexpected column title: "{0}" != "Used"!'
             .format(title))
    title = title_line[separator_cols[2] : separator_cols[3]].strip()
    if title != 'Available':
        warn('Unexpected column title: "{0}" != "Available"!'
             .format(title))
    title = title_line[separator_cols[3] : separator_cols[4]].strip()
    if title != 'Capacity':
        warn('Unexpected column title: "{0}" != "Capacity"!'
             .format(title))
    title = title_line[separator_cols[4] : ].strip()
    if title != 'Mounted on':
        warn('Unexpected column title: "{0}" != "Mounted on"!'
             .format(title))

    # create result
    result = []
    for line in out[1:]:
        stats = FilesystemFillState()
        stats.name = line[ : separator_cols[0]].strip()
        stats.size = int(line[separator_cols[0] : separator_cols[1]].strip()) \
                     * DF_SIZE_UNIT
        stats.used = int(line[separator_cols[1] : separator_cols[2]].strip()) \
                     * DF_SIZE_UNIT
        stats.available = int(line[separator_cols[2] : separator_cols[3]]\
                              .strip()) * DF_SIZE_UNIT
        stats.capacity = int(line[separator_cols[3] : separator_cols[4]]\
                             .strip()[:-1])
        stats.mount_point = line[separator_cols[4] : ].strip()

        # more checks: does match capacity
        capacity = 100. * stats.used / stats.size
        if abs(capacity - stats.capacity) > 5.:
            warn('capacities for {0} deviate more than 5%: '
                 '{1} != {2:.2f}(={3}/{4})'.format(
                 stats.name, stats.capacity, capacity, stats.used, stats.size))

        size = stats.used + stats.available
        if float(abs(stats.size - size)) / float(max(stats.size, size)) > 0.1:
            warn('size for {0} differs by more than 10% from used+available!'
                 .format(stats.name))

        result.append(stats)

    return result


def get_all_statvfs_fills(*mounts):
    """ run get_fill_from_statvfs on given MountPoints or get_all_mounts """

    if not mounts:
        mounts = get_all_mounts()

    return [get_fill_from_statvfs(mount) for mount in mounts]


def get_fill_from_statvfs(path):
    """ wrapper around os.statvfs

    :param path: a MountPoint, a path string or (since python3.3) a (os-level?)
                 file descriptor
    :returns: FilesystemFillState

    Tries to give results compatible with those from df_wrapper

    runs os.statvfs, but actually only uses f_blocks, f_frsize, f_bsize,
    f_bavail of result to set values for fields size and available fields of
    :py:class:`FilesystemFillState`. Fields used and capacity are calculated
    from that.

    If given a path or file descriptor, field name is set to given arg path,
    mount_point is not set (but can easily determined from path using
    :py:func:`find_mount_point`).

    If given a MountPoint, name is set to path.spec and mount_point to
    path.file
    """

    result = FilesystemFillState()
    if isinstance(path, MountPoint):
        stat_struct = os.statvfs(path.file)
        result.name = path.spec
        result.mount_point = path.file
    else:
        stat_struct = os.statvfs(path)
        result.name = path
        # result.mount_point is not set

    # fields according to "man statvfs":
    # unsigned long  f_bsize;    /* Filesystem block size */
    # unsigned long  f_frsize;   /* Fragment size */
    # fsblkcnt_t     f_blocks;   /* Size of fs in f_frsize units */
    # fsblkcnt_t     f_bfree;    /* Number of free blocks */
    # fsblkcnt_t     f_bavail;   /* Number of free blocks for
    #                               unprivileged users */
    # fsfilcnt_t     f_files;    /* Number of inodes */
    # fsfilcnt_t     f_ffree;    /* Number of free inodes */
    # fsfilcnt_t     f_favail;   /* Number of free inodes for
    #                               unprivileged users */
    # unsigned long  f_fsid;     /* Filesystem ID */
    # unsigned long  f_flag;     /* Mount flags */
    # unsigned long  f_namemax;  /* Maximum filename length */
    # (f_fsid is not included in result on linux)

    result.size = stat_struct.f_blocks * stat_struct.f_frsize
    result.available = stat_struct.f_bavail * stat_struct.f_bsize
    result.used = result.size - result.available
    if result.size == 0:
        result.capacity = 0
    else:
        result.capacity = float(result.used) / float(result.size) * 100.

    return result


class MountPoint:
    """ a mount point as in /proc/mounts a.k.a /etc/fstab

    (field names taken from man fstab)
    """
    def __init__(self):
        self.spec = None
        self.file = None
        self.vfstype = None
        self.options = None
        self.freq = None
        self.passno = None

    def __str__(self):
        if self.spec is None and self.file is None and self.type is None:
            return '[MountPoint uninitialized]'
        else:
            return '[MountPoint {0.spec} at {0.file} (type {0.vfstype})]' \
                   .format(self)

MOUNT_REGEXP = r'^\s*(?P<spec>.+)' \
                 '\s+(?P<file>.+)' \
                 '\s+(?P<vfstype>.+)' \
                 '\s+(?P<options>.+)' \
                 '\s+(?P<freq>\d+)' \
                 '\s+(?P<passno>\d+)\s*$'

def get_all_mounts():
    """ parse /proc/mounts

    does not return those with type in :py:data:IGNORE_MOUNT_TYPES

    :results: list of :class:`MountPoint`
    """
    result = []
    with open('/proc/mounts', 'rt') as file_handle:
        for line in file_handle:
            parts = line.split()
            matches = re.match(MOUNT_REGEXP, line)
            if not matches:
                raise ValueError('failed to interpret mount line "{0}"!'
                                 .format(line))
            new_mount = MountPoint()
            for field_name, value in matches.groupdict().items():
                setattr(new_mount, field_name, value)
            if new_mount.vfstype in IGNORE_MOUNT_TYPES:
                continue
            result.append(new_mount)

    return result


def find_mount_point(path):
    """ repeat os.ismount of parent's parent's parent... until returns true

    taken from, answer by larsmans from Dec 15 2010 on
    http://stackoverflow.com/questions/4453602/how-to-find-the-mountpoint-a-file-resides-on
    """
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path
